/*
 * Solaris kernel module to make a large contiguous buffer available to
 * user space. Written with the mga driver in the Utah GLX project in mind.
 *
 * Copyright (C) 2000 Andreas Ehliar <ehliar@lysator.liu.se>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * ANDREAS EHLIAR, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *
 *
 * Well, I can't say that I'm that fond of the Solaris DDI, but
 * I guess you might get used to it after a while.
 *
 * Since I'm no Solaris wizard I'm not sure that this is the most
 * efficient way to allocate a large contiguous buffer.
 * 
 * I'm using kmem_alloc() to allocate the DMA buffer. Unfortunately you
 * will most likely waste quite a lot of pages. Typically only about
 * 7 MB would be contiguous out of 8 MB if I forceloaded the module
 * in /etc/system
 *
 * Since it is hard to get contiguous memory after a while the module
 * will fail to unload unless allow_detach is set. This can be set with the
 * PHYSMAP_ALLOWDETACH ioctl.
 *
 * Only tested on Solaris 7/X86 on a dual celeron 466.
 *
 *   / Andreas Ehliar - 2000-03-30
 * 
 *
 * References:
 *
 * Writing Device Drivers Answerbook
 *    Available at http://docs.sun.com/
 *
 * Sample drivers in the SUNWdrvs package
 *    http://soldc.sun.com/developer/support/driver/src/SUNWdrvs.html
 *
 * Various man pages from section 9
 *
 * Driver Development FAQ
 *    http://soldc.sun.com/developer/support/driver/faqs.html
 */

#include <sys/modctl.h>
#include <sys/devops.h>
#include <sys/conf.h>
#include <sys/sunddi.h>
#include <sys/stat.h>
#include <sys/open.h>
#include <sys/file.h>
#include <sys/kmem.h>
#include <sys/ddi.h>
#include "physmap.h"

static int physmap_attach(dev_info_t *dip,ddi_attach_cmd_t cmd);
static int physmap_detach(dev_info_t *dip,ddi_attach_cmd_t cmd);
static int physmap_open(dev_t *devp, int flag, int otyp, cred_t *credp);
static int physmap_close(dev_t dev, int flag, int otyp, cred_t *credp);
static int physmap_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg, void **result);
static int physmap_ioctl(dev_t dev, int cmd, intptr_t arg, int mode, cred_t *credp, int *rvalp);

static struct cb_ops physmap_cb_ops = {
  physmap_open, /* open */ 
  physmap_close, /* close */
  nodev, /* strategy */
  nodev, /* print */
  nodev, /* dump */
  nodev, /* read */
  nodev,/* write */
  physmap_ioctl, /* ioctl */
  nodev, /* devmap */
  nodev, /* mmap */
  nodev, /* segmap */
  nochpoll,
  ddi_prop_op,
  NULL,
  D_MP,
  CB_REV,
  nodev,
  nodev
};
  

static struct dev_ops physmap_ops = {
  DEVO_REV,
  0,
  physmap_getinfo,
  nulldev,
  nulldev,
  physmap_attach,
  physmap_detach,
  nodev,
  &physmap_cb_ops,
  NULL
};

/* Well, I'm not completely sure about this one.
 * I guess there might be problems getting it to work on a SPARC. */
static ddi_dma_attr_t physmap_dma_attr = {
  DMA_ATTR_V0,
  (uint64_t)0,
  (uint64_t)0xffffffff,
  (uint64_t)0xffffffff,
  (uint64_t)1,
  0xf, /* not sure about the burstsize */
  1, /* not sure about this either */
  (uint64_t)0xffffffff,
  (uint64_t)0xffffffff,
  1, /* no scatter gather */
  1,
  0
 
};
  
static int initialized=0; /* Has the driver has been attached? */

static int physmap_instance=0; /* We will only allow one instance for now */
static dev_info_t *physmap_dip;

/* size of aperture we are trying to map, default 8 MB */

static uint32_t regionsize=8*1024*1024; 
static uint32_t largest_slice=0;
static uint32_t slice_phys=0;
static caddr_t dmaregion=NULL;

static int allow_detach=0; /* By default we won't allow the module to be unloaded
			    * since it would be hard to get a contiguous memory
			    * region after the computer has been used for a while.
			    */

static ddi_dma_cookie_t	dmacookie;
static ddi_dma_handle_t physmap_dmahandle;

/* Assume 32 bit address space. (I can't test this on an Ultra myself unfortunately) */
struct memory_slices{
  uint32_t addr;
  uint32_t size;
};


/* Merge the slices ddi_dma_nextcookie() returned into larger regions
 * I'm certain this could be made more efficient but it is only run once
 * so it doesn't really matter if it isn't that fast. */

static int mergeslices(struct memory_slices *slices,int size)
{
  int i,j;
  int newsize=size; /* newsize represents the amount of remaining slies */

  /* First we try to find the buddy of the last slice */
  for(i=size;i>0;i--){
    j=0;
    /* Search forward until we find it or the last element */
    while((slices[i].addr != slices[j].size+slices[j].addr)  && (j < i)){
      j++;
    }
    
    if(j!=i){
      slices[j].size+=slices[i].size; /* Merge our element with its buddy */

      /* XXX - There was a good reason for these lines but can't remember
       * it right now */
      slices[i].size=slices[newsize].size;
      slices[i].addr=slices[newsize].addr;
      newsize--;
    }
  }

  return newsize;
}


static int physmap_attach(dev_info_t *dip,ddi_attach_cmd_t cmd)
{
  ddi_device_acc_attr_t dev_attr;
  uint_t count;
  ddi_dma_attr_t tmp_dma_attr;



  if(initialized){
    /* We only allow this driver to be initialized once */
    return DDI_FAILURE;
  }

    
  initialized=PHYSMAP_INIT_ATTACHING; /* possible race? */

  tmp_dma_attr=physmap_dma_attr; /* Done this way in the device driver manual */

  physmap_instance = ddi_get_instance(dip);

  dev_attr.devacc_attr_version = DDI_DEVICE_ATTR_V0;
  dev_attr.devacc_attr_endian_flags = DDI_NEVERSWAP_ACC;
  dev_attr.devacc_attr_dataorder = DDI_MERGING_OK_ACC; /* Write combining? */


  if(cmd==DDI_ATTACH){

    /* Some magic to get an area we can do DMA to. Don't blame me, I didn't
     * design Solaris this way. */

    if(ddi_dma_alloc_handle(dip,&tmp_dma_attr,
			    DDI_DMA_SLEEP,NULL,&physmap_dmahandle)==DDI_SUCCESS){

      initialized=PHYSMAP_INIT_HANDLE;

      /* kmem_alloc() will typically return contiguous memory regions if
       * called early after boot according to 
       * http://soldc.sun.com/developer/support/driver/faqs.html#NQ5.1
       */
      dmaregion=kmem_alloc(regionsize,KM_SLEEP);
      if(dmaregion){

	initialized=PHYSMAP_INIT_KMEM_ALLOC;

	/* We can use ddi_dma_addr_bind_handle() for memory allocated by kmem_alloc()
	 * according to the URL referenced above */
	if(ddi_dma_addr_bind_handle(physmap_dmahandle,NULL,dmaregion,regionsize,
					      DDI_DMA_RDWR|DDI_DMA_CONSISTENT,
					      DDI_DMA_SLEEP,NULL,&dmacookie,&count)==DDI_DMA_MAPPED){
	  ddi_dma_cookie_t mycookie;
	  struct memory_slices *slices=NULL;
	  int n=0,i=0;
	  initialized=PHYSMAP_INIT_BOUND;

	  /* Assume page size is 4096 */
	  slices=kmem_zalloc(regionsize/4096*sizeof(struct memory_slices),KM_SLEEP);

	  if(slices){
	    int currpos;

	    /* Collect physical addresses of the various parts of dmaregion */
	    slices[0].addr = dmacookie.dmac_address;
	    slices[0].size = dmacookie.dmac_size;
	    currpos=0;

	    while(--count > 0){
	      ddi_dma_nextcookie(physmap_dmahandle,&mycookie);

	      currpos++;
	      slices[currpos].addr=mycookie.dmac_address;
	      slices[currpos].size=mycookie.dmac_size;
	    }


	    /* merge the slices returned by ddi_dma_nextcookie
	     * and find the largest slice */

	    n = mergeslices(slices,currpos);
	    for(i=0;i<=n;i++){
	      if(slices[i].size > largest_slice){
		largest_slice=slices[i].size;
		slice_phys=slices[i].addr;
	      }
	    }
	    kmem_free(slices,regionsize/4096*sizeof(struct memory_slices));
	    
	    if(ddi_create_minor_node(dip,"physmap",S_IFCHR,0,DDI_NT_DISPLAY,0)==DDI_SUCCESS){
	      initialized=PHYSMAP_INIT_DONE;
	      return DDI_SUCCESS;
	    }
	  }
	}

      }
    }

    cmn_err(CE_WARN,"physmap: Cannot allocate DMA resources\n");
    physmap_detach(dip,DDI_DETACH);
  }

  return DDI_FAILURE;
}


static int physmap_detach(dev_info_t *dip,ddi_attach_cmd_t cmd)
{
  if(!initialized){
    /* I guess this can't happen */
    cmn_err(CE_WARN,"physmap: physmap_detach called without an initialized device");
    return DDI_FAILURE;
  }

  /* Might not want to detach this interface in the event of OOM since it would
   * be hard to get a contiguous region again */
  if(!allow_detach){
    return DDI_FAILURE;
  }

  /* XXX - return values? */

  if(cmd==DDI_DETACH){
    switch(initialized){
    case PHYSMAP_INIT_DONE:
      ddi_remove_minor_node(dip,"physmap");
    case PHYSMAP_INIT_BOUND:
      ddi_dma_unbind_handle(physmap_dmahandle);
    case PHYSMAP_INIT_KMEM_ALLOC:
      kmem_free(dmaregion,regionsize);
    case PHYSMAP_INIT_HANDLE:
      ddi_dma_free_handle(&physmap_dmahandle);
    case PHYSMAP_INIT_ATTACHING:
      break;
      
    default:
      cmn_err(CE_WARN,"physmap: Something went horribly wrong, initialized is %d\n",initialized);
    }
    initialized=0;
    return DDI_SUCCESS;
  }

  return DDI_FAILURE;
}

static struct modldrv modldrv = {
  &mod_driverops,
  "physmap driver v0.9",
  &physmap_ops
};

static struct modlinkage modlinkage = {
    MODREV_1,
    {&modldrv,
     NULL} /* No braces in the device driver manual, gcc didn't like that though */
};



static int physmap_open(dev_t *devp, int flag, int otyp, cred_t *credp)
{
  if(getminor(*devp)!=0){
    cmn_err(CE_WARN,"physmap: physmap_open called with minor!=0"); /* Shouldn't happen */
    return EINVAL;
  }

  if(!initialized){
    return ENXIO;
  }

  if(otyp!=OTYP_CHR){
    return ENXIO;
  }
  
  if(flag == FEXCL){
    return EINVAL;
  }

  if (drv_priv(credp) == EPERM){
    /* Well, can't say that I really understand this */
     return (EPERM);
  }

  return 0;
}

static int 
physmap_ioctl(dev_t dev, int cmd, intptr_t arg, int mode, cred_t *credp, int *rvalp)
{
  struct physmap_info info;

  switch(cmd){
  case PHYSMAP_GETINFO:
    info.largest_slice = largest_slice;
    info.slice_phys = slice_phys;
    info.regionsize = regionsize;
    info.allow_detach = allow_detach;
    if(ddi_copyout(&info,(void *)arg,sizeof(info),mode)){
      return EFAULT;
    }
    return 0;
  case PHYSMAP_ALLOWDETACH:
    allow_detach=1;
    return 0;
  case PHYSMAP_DENYDETACH:
    allow_detach=0;
    return 0;
  default:
    return ENOTTY; /* As used in the device driver sample source */
  }
}
    

int physmap_close(dev_t dev, int flag, int otyp, cred_t *credp)
{
  /* No sanity checks for now */
  return 0;
}

/* Pretty much copied from getinfo(9E) */
static int
physmap_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg, void **result)
{
  int error;

  if(!initialized){
    return DDI_FAILURE;
  }

  switch(infocmd){
  case DDI_INFO_DEVT2DEVINFO:
    *result = physmap_dip;
    error = DDI_SUCCESS;
    
    break;
  case DDI_INFO_DEVT2INSTANCE:
    *result = (void *)physmap_instance;
    error = DDI_SUCCESS;
    break;
  default:
    error = DDI_FAILURE;
  }
  
  return (error);
}


int _init(void)
{
  int error;
  initialized=0;

  error=mod_install(&modlinkage);

  if(error){
    cmn_err(CE_WARN,"physmap: Cannot attach. mod_install returned %d\n",error);
    return error;
  }
  
  dmaregion=kmem_alloc(regionsize,KM_SLEEP);
  if(!dmaregion){
    cmn_err(CE_WARN,"physmap: Cannot allocate 0x%x bytes of memory\n",regionsize);
    mod_remove(&modlinkage);
    return DDI_FAILURE;
  }

  return DDI_SUCCESS;
}

int _fini(void)
{
  int error;
  error=mod_remove(&modlinkage);

  if(error){
    return error;
  }

  return DDI_SUCCESS;
}

int _info(struct modinfo *modinfop)
{
  return (mod_info(&modlinkage, modinfop));
}
